# Print out, for each class of quartic twist parameter, the
# conjectural rank parity of a curve in that class which satisfies
# legendre(-1, S(d)) == 1.  Curves in the class which have
# legendre(-1, S(d)) == -1 have the opposite parity.
# This function puts the results in a big table, with units varying
# across rows and powers of 2 varying across columns.
def classifyQuarticTwistsOverQTable(sign = 1):
    K.<y> = NumberField(x-1)
    R = K.maximal_order()
    
    # for K = QQ, m_2 = 4
    #units2ToMvMod4 = getResiduesModPower(K.ideal(2)^4, 4)
    units2ToMvMod4 = [1, 3, 5, 7, 9, 11, 13, 15]
    print("Residues: " + str(units2ToMvMod4))
    
    for ord2 in range(0, 4):
        sys.stdout.write(str(ord2) + "&& ")
        sys.stdout.flush()
        for unit2Residue in units2ToMvMod4:
            twistParam = 2^ord2 * ((16 + sign*unit2Residue) % 16) * sign
            if (twistParam == 0): continue
            legendreCond = legendreCondition(K, twistParam)
            ETwist = EllipticCurve(QQ, [twistParam, 0])
            rankParity = getConjecturalRankParity(ETwist)
            #rankParity = ETwist.analytic_rank() % 2
            if legendreCond == 1:
                sys.stdout.write(str(rankParity))
            else:
                sys.stdout.write(str(1 - rankParity))
            if not (unit2Residue == units2ToMvMod4[-1]):
                sys.stdout.write("&")
            sys.stdout.flush()
        sys.stdout.write(" \\\\ \\hline\n")


def getResiduesModPower(ideal, power):
    residues = list(ideal.invertible_residues())
    residuesToPower = []
    for elem in residues:
        if ideal.reduce(elem^power) not in residuesToPower:
            residuesToPower.append(ideal.reduce(elem^power))
    return list(ideal.invertible_residues_mod(residuesToPower))


# Returns legendre(-1, S(d)), in the notation of the main thm.
# Right now this only works when 2 is stationary in K
def legendreCondition(K, d):
    ans = 1
    factors = K.factor(d)
    for factor in factors:
        if factor[1] % 4 == 2:
            if factor[0] != 2:
                ans *= factor[0].residue_symbol(-1, 2)
    return ans


def getConjecturalRankParity(E):
    twoSelmerRankParity = E.simon_two_descent()[1] % 2
    torsion = E.torsion_subgroup()
    twoTorsionCount = 0
    for g in torsion:
        if g + g == 0:
            twoTorsionCount += 1
    twoTorsionRank = log(twoTorsionCount, 2)
    return (twoSelmerRankParity + twoTorsionRank) % 2


classifyQuarticTwistsOverQTable(1)
classifyQuarticTwistsOverQTable(-1)
